import { JsonPipe } from '@angular/common';
import { Component, Inject, Input, OnChanges, OnInit, SecurityContext, SimpleChanges } from '@angular/core';
import { FormControl, FormGroup, ValidatorFn, Validators } from '@angular/forms';
import { MatDialog, MatDialogRef, MAT_DIALOG_DATA } from '@angular/material/dialog';
import { MatSelectChange } from '@angular/material/select';
import { MatSnackBar } from '@angular/material/snack-bar';
import { ActivatedRoute, Router } from '@angular/router';
import { KeyValue, SimpleResource, WfConf, WfHistoryEntry, WfParam, WfProcessStatus, WfSection } from '../common/is.model';
import { ISService } from '../common/is.service';
import { ResMetadataDialog } from '../resources/resources.component';
import { MatTableDataSource } from '@angular/material/table';
import { WfHistoryDialog } from '../wf-history/wf-history.component';

@Component({
	selector: 'app-wf-confs',
	templateUrl: './wf-confs.component.html',
	styleUrls: ['./wf-confs.component.css']
})
export class WfConfsComponent implements OnInit {
	section?: WfSection;
	confs: KeyValue[] = [];
	conf?: WfConf;

	constructor(public service: ISService, public route: ActivatedRoute, public router: Router, public dialog: MatDialog, public snackBar: MatSnackBar) {
	}

	ngOnInit() {
		this.route.params.subscribe(params => {
			let sectionId = params['section'];
			let confId = params['conf'];

			if (confId) {
				this.service.loadWfConfiguration(confId, (conf: WfConf) => {
					this.conf = conf;
					if (conf.section) {
						this.service.loadWfSections((data: WfSection[]) => this.section = data.find(s => s.id == conf?.section));
						this.service.loadWfConfigurations(conf.section, (data: KeyValue[]) => this.confs = data);
					}
				});
			} else if (sectionId) {
				this.service.loadWfSections((data: WfSection[]) => this.section = data.find(s => s.id == sectionId));
				this.service.loadWfConfigurations(sectionId, (data: KeyValue[]) => {
					this.confs = data;
					if (data.length > 0) {
						this.router.navigate(['/wfs/conf', data[0].k]);
					}
				});
			} else {
				console.log("One of the following parameters is missing: sectionId or confId");
			}
		});
	}

	openAddWfConfDialog(): void {
		const dialogRef = this.dialog.open(WfConfDialog, {
			data: {
				id: '',
				name: '',
				section: this.section?.id,
				enabled: true,
				priority: 75,
				workflow: '',
				schedulingEnabled: false,
				cronExpression: '0 30 12 1/1 * ?',
				cronMinInterval: 9600,
				details: new Map,
				configured: true,
				systemParams: new Map,
				userParams: new Map
			},
			width: '80%'
		});

		dialogRef.afterClosed().subscribe(result => {
			if (result) this.router.navigate(['/wfs/conf', result.id]);;
		});
	}
}

@Component({
	selector: 'wf-conf-dialog',
	templateUrl: 'wf-conf-dialog.html',
	styleUrls: ['./wf-confs.component.css']
})
export class WfConfDialog implements OnInit {

	wfTemplates: SimpleResource[] = [];
	wfParameters: WfParam[] = [];

	wfConfFormStep1 = new FormGroup({
		workflow: new FormControl('', [Validators.required]),
	});

	wfConfFormStep2 = new FormGroup({
		name: new FormControl('', [Validators.required]),
		//details: Map<string, string>,
		enabled: new FormControl(true, [Validators.required]),
		priority: new FormControl(75, [Validators.required, Validators.min(1), Validators.max(100)]),
	});

	wfConfFormStep3 = new FormGroup({
		//systemParams: Map<string, string>,
		//userParams: Map<string, string>
	});

	wfConfFormStep4 = new FormGroup({
		schedulingEnabled: new FormControl(false, [Validators.required]),
		cronExpression: new FormControl("", [Validators.required]),
		cronMinInterval: new FormControl("", [Validators.required]),
	});

	wfConfFormFinal = new FormGroup({});

	constructor(public dialogRef: MatDialogRef<ResMetadataDialog>, @Inject(MAT_DIALOG_DATA) public data: any, public service: ISService) {
		this.wfConfFormStep1.get('workflow')?.setValue(data.workflow);

		this.wfConfFormStep2.get('name')?.setValue(data.name);
		this.wfConfFormStep2.get('enabled')?.setValue(data.enabled);
		this.wfConfFormStep2.get('priority')?.setValue(data.priority);
		//details
		//systemParams,
		//userParams
		this.wfConfFormStep4.get('schedulingEnabled')?.setValue(data.schedulingEnabled);
		this.wfConfFormStep4.get('cronExpression')?.setValue(data.cronExpression);
		this.wfConfFormStep4.get('cronMinInterval')?.setValue(data.cronMinInterval);
	}
	ngOnInit(): void {
		if (this.data.id && this.data.workflow) {
			this.prepareWfParameters(this.data.workflow);
		}
		this.service.loadSimpleResources("wf_template", (data: SimpleResource[]) => this.wfTemplates = data);
	}

	onChangeWfTemplate(e: MatSelectChange): void {
		this.prepareWfParameters(e.value);
	}

	prepareWfParameters(wfTemplateId: string): void {
		this.service.loadSimpleResourceContent(wfTemplateId, (data: any) => {
			console.log(data);
			if (data.parameters) {
				this.wfParameters = data.parameters;
			} else {
				this.wfParameters = JSON.parse(data).parameters;
			}

			this.wfConfFormStep3.controls = {};

			this.wfParameters.forEach(p => {
				let validations: ValidatorFn[] = [];
				if (p.required) { validations.push(Validators.required); }
				if (p.type == 'number') { validations.push(Validators.pattern('^[0-9]*$')); }
				this.wfConfFormStep3.addControl(p.name, new FormControl(this.data.userParams[p.name], validations));
			})
		});
	}

	onSubmit(): void {
		const conf = Object.assign({}, this.data, this.wfConfFormStep1.value, this.wfConfFormStep2.value, this.wfConfFormStep4.value);
		conf.details = {};
		conf.systemParams = {};
		conf.userParams = this.wfConfFormStep3.value;

		this.service.saveWfConfiguration(conf, (data: void) => this.dialogRef.close(data), this.wfConfFormFinal);
	}

	onNoClick(): void {
		this.dialogRef.close();
	}
}

@Component({
	selector: 'wf-conf-single',
	templateUrl: 'wf-conf-single.html',
	styleUrls: ['./wf-confs.component.css']
})
export class WfConfSingle implements OnInit, OnChanges {

	@Input() conf?: WfConf;
	prevConfId = '';

	historyDatasource: MatTableDataSource<WfHistoryEntry> = new MatTableDataSource<WfHistoryEntry>([]);
	colums: string[] = ['processId', 'status', 'startDate', 'endDate'];

	constructor(public service: ISService, public dialog: MatDialog, public snackBar: MatSnackBar) { }

	ngOnInit(): void {
		if (this.conf) {
			this.service.loadWfHistoryForConf(this.conf?.id, (data: WfHistoryEntry[]) => this.historyDatasource.data = data);
		}
	}

	ngOnChanges(changes: SimpleChanges): void {
		if (this.conf && this.conf.id != this.prevConfId) {
			this.prevConfId = this.conf.id;
			this.service.loadWfHistoryForConf(this.conf?.id, (data: WfHistoryEntry[]) => this.historyDatasource.data = data);
		}
	}



	launchWfConf() {
		if (this.conf?.id && this.conf?.workflow) {
			this.service.startWfConfiguration(this.conf?.id, (data: WfProcessStatus) => this.snackBar.open('Workflow launched !!!', 'INFO', { duration: 5000 }));
		}
	}

	editConf() {
		const dialogRef = this.dialog.open(WfConfDialog, {
			data: this.conf,
			width: '80%'
		});
	}

	deleteConf() {
		if (this.conf?.destroyWf) {
			if (this.conf?.id && this.conf?.workflow) {
				this.service.startDestroyWfConfiguration(this.conf?.id, (data: WfProcessStatus) => this.snackBar.open('Destroy Workflow launched, PLEASE WAIT !!!', 'INFO', { duration: 5000 }));
			}
		} else if (this.conf?.id) {
			this.service.deleteWfConfiguration(this.conf?.id, (data: void) => {
				this.snackBar.open('Configuration deleted !!!', 'INFO', { duration: 5000 });
				this.conf = undefined;
			});
		}
	}

	openWfHistoryDialog(wf: WfHistoryEntry): void {
		const wfDialogRef = this.dialog.open(WfHistoryDialog, {
			data: wf
		});
	}

}
