package eu.dnetlib.data.collect;

import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.nio.charset.StandardCharsets;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Queue;
import java.util.Set;
import java.util.concurrent.PriorityBlockingQueue;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import java.util.stream.StreamSupport;

import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClientBuilder;
import org.dom4j.Document;
import org.dom4j.DocumentException;
import org.dom4j.DocumentHelper;
import org.dom4j.Node;

import eu.dnetlib.common.utils.xml.XmlCleaner;
import eu.dnetlib.dsm.model.ApiParam;

public class OaiCollector implements DnetCollector {

	@Override
	public String protocol() {
		return "OAI";
	}

	@Override
	public Stream<String> collect(final String baseUrl, final Set<ApiParam> params) {
		return collect(baseUrl, params, null, null);
	}

	@Override
	public Stream<String> collect(final String baseUrl, final Set<ApiParam> params, final LocalDateTime from, final LocalDateTime until) {
		final String mdFormat = params.stream()
			.filter(p -> p.getParam().equals("format"))
			.map(p -> p.getValue())
			.findFirst()
			.orElse("oai_dc");

		final String setParam = params.stream()
			.filter(p -> p.getParam().equals("set"))
			.map(p -> p.getValue())
			.findFirst()
			.orElse("");

		final List<String> sets = Arrays.stream(setParam.split(","))
			.map(String::trim)
			.filter(StringUtils::isNotBlank)
			.distinct()
			.collect(Collectors.toList());

		if (baseUrl == null || baseUrl.isEmpty()) { throw new RuntimeException("Param 'baseurl' is null or empty"); }

		if (mdFormat == null || mdFormat.isEmpty()) { throw new RuntimeException("Param 'mdFormat' is null or empty"); }

		if (sets.isEmpty()) {
			// If no set is defined, ALL the sets must be harvested
			return this.createStream(baseUrl, null, mdFormat, from, until);
		} else {
			return sets.stream()
				.map(set -> this.createStream(baseUrl, set, mdFormat, from, until))
				.flatMap(o -> o);
		}

	}

	private Stream<String> createStream(final String baseUrl,
		final String set,
		final String mdFormat,
		final LocalDateTime from,
		final LocalDateTime until) {
		final Iterable<String> iter = () -> new OaiIterator(baseUrl, set, mdFormat, from, until);
		return StreamSupport.stream(iter.spliterator(), false);
	}

	public class OaiIterator implements Iterator<String> {

		private final Queue<String> queue = new PriorityBlockingQueue<>();
		private final DateTimeFormatter oaiDateFormatter = DateTimeFormatter.ofPattern("yyyy-MM-dd");
		private final String baseUrl;
		private final String set;
		private final String mdFormat;
		private final LocalDateTime from;
		private final LocalDateTime until;

		private String token;
		private boolean started;

		public OaiIterator(final String baseUrl, final String mdFormat, final String set, final LocalDateTime from, final LocalDateTime until) {
			this.baseUrl = baseUrl;
			this.mdFormat = mdFormat;
			this.set = set;
			this.from = from;
			this.until = until;
			this.started = false;
		}

		private void verifyStarted() {
			if (!this.started) {
				this.started = true;
				this.token = firstPage();
			}
		}

		@Override
		public boolean hasNext() {
			synchronized (queue) {
				verifyStarted();
				return !queue.isEmpty();
			}
		}

		@Override
		public String next() {
			synchronized (queue) {
				verifyStarted();
				final String res = queue.poll();
				while (queue.isEmpty() && token != null && !token.isEmpty()) {
					token = otherPages(token);
				}
				return res;
			}
		}

		@Override
		public void remove() {}

		private String firstPage() {
			try {
				String url = baseUrl + "?verb=ListRecords&metadataPrefix=" + URLEncoder.encode(mdFormat, "UTF-8");
				if (set != null && !set.isEmpty()) {
					url += "&set=" + URLEncoder.encode(set, StandardCharsets.UTF_8);
				}

				if (from != null) {
					url += "&from=" + URLEncoder.encode(from.format(oaiDateFormatter), StandardCharsets.UTF_8);
				}
				if (until != null) {
					url += "&until=" + URLEncoder.encode(until.format(oaiDateFormatter), StandardCharsets.UTF_8);
				}

				return downloadPage(url);
			} catch (final UnsupportedEncodingException e) {
				throw new RuntimeException(e);
			}
		}

		private String otherPages(final String resumptionToken) {
			try {
				return downloadPage(baseUrl + "?verb=ListRecords&resumptionToken=" + URLEncoder.encode(resumptionToken, "UTF-8"));
			} catch (final UnsupportedEncodingException e) {
				throw new RuntimeException(e);
			}
		}

		private String downloadPage(final String url) {
			try (final CloseableHttpClient httpClient = HttpClientBuilder.create().build()) {
				final HttpGet request = new HttpGet(url);

				try (final CloseableHttpResponse response = httpClient.execute(request)) {
					final String xml = IOUtils.toString(response.getEntity().getContent(), StandardCharsets.UTF_8);

					final Document doc = tryParseXml(xml);

					final Node errorNode = doc.selectSingleNode("/*[local-name()='OAI-PMH']/*[local-name()='error']");
					if (errorNode != null) {
						final String code = errorNode.valueOf("@code");
						if ("noRecordsMatch".equalsIgnoreCase(code.trim())) {
							return null;
						} else {
							throw new RuntimeException(code + " - " + errorNode.getText());
						}
					}
					long added_records = 0;
					for (final Node n : doc.selectNodes("//*[local-name()='ListRecords']/*[local-name()='record']")) {
						queue.add(n.asXML());
						added_records++;
					}
					if (added_records == 0) { throw new RuntimeException("OAI Iterator Exception: Empty records list for " + url); }

					return doc.valueOf("//*[local-name()='resumptionToken']");
				}
			} catch (final Exception e) {
				throw new RuntimeException(e);
			}
		}

		private Document tryParseXml(final String xml) {
			try {
				return DocumentHelper.parseText(xml);
			} catch (final DocumentException e) {
				final String cleaned = XmlCleaner.cleanAllEntities(xml);
				try {
					return DocumentHelper.parseText(cleaned);
				} catch (final DocumentException e1) {
					throw new RuntimeException(e);
				}
			}
		}
	}
}
