package eu.dnetlib.manager.wf.cron;

import java.time.LocalDateTime;
import java.time.temporal.ChronoUnit;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.scheduling.support.CronExpression;
import org.springframework.stereotype.Service;

import eu.dnetlib.manager.history.WorkflowLogger;
import eu.dnetlib.manager.wf.WorkflowManagerService;
import eu.dnetlib.manager.wf.model.WorkflowConfiguration;
import eu.dnetlib.manager.wf.repository.WorkflowConfigurationRepository;
import eu.dnetlib.manager.wf.workflows.procs.ProcessRegistry;
import eu.dnetlib.manager.wf.workflows.procs.WorkflowProcess;

@Service
public class ScheduledWorkflowLauncher {

	private static final Log log = LogFactory.getLog(ScheduledWorkflowLauncher.class);

	@Autowired
	private WorkflowManagerService wfManagerService;

	@Autowired
	private ProcessRegistry processRegistry;

	@Autowired
	private WorkflowConfigurationRepository workflowConfigurationRepository;

	@Autowired
	private WorkflowLogger logger;

	@Value("${dnet.workflow.scheduler.windowSize:1800000}")
	private int windowSize; // 1800000 are 30 minutes

	@Scheduled(fixedRateString = "${dnet.workflow.scheduler.fixedRate:900000}") // 900000 are 5 minutes
	public void verifySheduledWorkflows() {
		log.debug("Verifying scheduled workflows - START");

		workflowConfigurationRepository.findAll()
			.stream()
			.filter(WorkflowConfiguration::isEnabled)
			.filter(WorkflowConfiguration::isConfigured)
			.filter(WorkflowConfiguration::isSchedulingEnabled)
			.filter(this::isNotRunning)
			.filter(this::isReady)
			.forEach(conf -> {
				try {
					wfManagerService.startWorkflowConfiguration(conf);
				} catch (final Exception e) {
					log.error("Error launching scheduled wf conf: " + conf.getId(), e);
				}
			});
		log.debug("Verifying scheduled workflows - END");
	}

	private boolean isReady(final WorkflowConfiguration conf) {
		final LocalDateTime lastExecutionDate = calculateLastExecutionDate(conf.getId());

		final LocalDateTime now = LocalDateTime.now();

		final String cron = conf.getCronExpression();

		if (CronExpression.isValidExpression(cron)) {
			final int minInterval = conf.getCronMinInterval(); // in minutes

			final boolean res;
			if (lastExecutionDate != null) {
				final long elapsed = ChronoUnit.MINUTES.between(lastExecutionDate, now);
				res = elapsed > minInterval && verifyCron(cron, now);
			} else {
				res = verifyCron(cron, now);
			}

			if (log.isDebugEnabled()) {
				log.debug("**************************************************************");
				log.debug("WORKFLOW CONFIGURATION ID : " + conf.getId());
				log.debug("NOW                       : " + now);
				log.debug("LAST EXECUTION DATE       : " + lastExecutionDate);
				log.debug("MIN INTERVAL (minutes)    : " + minInterval);
				log.debug("WINDOW SIZE (ms)          : " + windowSize);
				log.debug("MUST BE EXECUTED          : " + res);
				log.debug("**************************************************************");
			}

			return res;
		}

		return false;
	}

	private LocalDateTime calculateLastExecutionDate(final String id) {
		return logger.getLastLogForConfiguration(id)
			.map(e -> e.getEndDate())
			.orElse(LocalDateTime.MIN);
	}

	private boolean verifyCron(final String cronExpression, final LocalDateTime now) {
		try {
			final CronExpression cron = CronExpression.parse(cronExpression);

			final LocalDateTime date = now.minus(windowSize, ChronoUnit.MINUTES);
			final LocalDateTime nextDate = cron.next(date);

			if (log.isDebugEnabled()) {
				log.debug("NEXT EXECUTION DATE: " + nextDate);
				log.debug("FIRED              : " + nextDate.isBefore(now));
			}
			return nextDate.isBefore(now);
		} catch (final Exception e) {
			log.error("Error calculating next cron event: " + cronExpression, e);
			return false;
		}
	}

	private boolean isNotRunning(final WorkflowConfiguration conf) {
		final WorkflowProcess p = processRegistry.findProcsByConfigurationId(conf.getId());

		if (p != null) {
			switch (p.getStatus()) {
			case created:
				return false;
			case running:
				return false;
			default:
				break;
			}
		}

		return true;
	}

}
