package eu.dnetlib.manager.wf.workflows.procs;

import java.time.LocalDateTime;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.PriorityBlockingQueue;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import eu.dnetlib.errors.WorkflowManagerException;
import eu.dnetlib.manager.wf.model.WorkflowConfiguration;
import eu.dnetlib.manager.wf.workflows.util.WorkflowsConstants;

@Service
public class ProcessRegistry {

	private static final Log log = LogFactory.getLog(ProcessRegistry.class);
	private final Map<String, WorkflowProcess> procs = new HashMap<>();
	private final Map<String, WorkflowProcess> byConfId = new HashMap<>();

	private final PriorityBlockingQueue<WorkflowProcess> pendingProcs = new PriorityBlockingQueue<>();

	@Value("${dnet.wf.registry.size:100}")
	private int maxSize;

	synchronized public int countRunningWfs() {
		int count = 0;
		for (final Map.Entry<String, WorkflowProcess> e : this.procs.entrySet()) {
			final WorkflowProcess proc = e.getValue();
			if (!proc.isTerminated()) {
				count++;
			}
		}
		return count;
	}

	public WorkflowProcess findProcess(final String procId) {
		return this.procs.get(procId);
	}

	public Collection<WorkflowProcess> listProcesses() {
		return this.procs.values();
	}

	public WorkflowProcess findProcsByConfigurationId(final String id) {
		return this.byConfId.get(id);
	}

	public void registerProcess(final WorkflowProcess process, final WorkflowConfiguration conf) throws WorkflowManagerException {
		if (this.procs.containsValue(process) || this.procs.containsKey(process.getId())) {
			log.error("Already registerd process: " + process);
			throw new WorkflowManagerException("Already registerd process: " + process);
		}

		if (this.procs.size() >= this.maxSize) {
			removeOldestProcess();
		}

		this.procs.put(process.getId(), process);
		this.byConfId.put(conf.getId(), process);

		synchronized (this.pendingProcs) {
			if (this.pendingProcs.size() > WorkflowsConstants.MAX_PENDING_PROCS_SIZE) {
				log.warn("Wf [" + process.getName() + "] not launched, Max number of pending procs reached: " + WorkflowsConstants.MAX_PENDING_PROCS_SIZE);
				throw new WorkflowManagerException("Max number of pending procs reached: " + WorkflowsConstants.MAX_PENDING_PROCS_SIZE);
			}
			this.pendingProcs.put(process);

			log.info("WorkflowProcess [" + process + "] in queue, priority=" + process.getPriority());
		}
	}

	private void removeOldestProcess() {
		LocalDateTime oldDate = LocalDateTime.now();
		String oldId = null;

		for (final Map.Entry<String, WorkflowProcess> e : this.procs.entrySet()) {
			final WorkflowProcess proc = e.getValue();

			if (proc.isTerminated()) {
				final LocalDateTime date = proc.getLastActivityDate();
				if (date.isBefore(oldDate)) {
					oldDate = date;
					oldId = e.getKey();
				}
			}
		}

		if (oldId != null) {
			unregisterProcess(oldId);
		}

	}

	public void unregisterProcess(final String procId) {
		synchronized (this) {
			final WorkflowProcess process = this.procs.remove(procId);
			if (process != null) {
				this.byConfId.entrySet()
					.stream()
					.filter(e -> e.getValue().getId().equals(process.getId()))
					.map(e -> e.getKey())
					.forEach(confId -> this.byConfId.remove(confId, process));
			}
		}

	}

	public WorkflowProcess nextProcessToStart() {
		synchronized (this.pendingProcs) {
			return this.pendingProcs.poll();
		}
	}

}
